/*
  miac.cpp - Library driver for the Matrix MIAC Industrial Controller
  Created by Leigh Morris (Matrix TSL) March 2016

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
*/

#include "miac.h"
#include "font610.h"

#define MIAC_LCD_DB_VAL_DDR(x)	{DDRC = 0xff; PORTC = (x);}
#define MIAC_LCD_DB_IN_DDR		{DDRC = 0;}
#define MIAC_LCD_DB_GET			(PINC)

#define MIAC_I1_IN_DDR			{DDRG &= 0xfe;}
#define MIAC_I1_GET				(PING & 1)
#define MIAC_I2_IN_DDR			{DDRG &= 0xfd;}
#define MIAC_I2_GET				(PING & 2)

#define MIAC_LCD_EN_CLR			{PORTD &= ~(1<< 6);}	
#define MIAC_LCD_EN_SET			{PORTD |= (1<< 6);}	
#define MIAC_LCD_RS_CLR			{PORTD &= ~(1<< 4);}	
#define MIAC_LCD_RS_SET			{PORTD |= (1<< 4);}	
#define MIAC_LCD_RW_CLR			{PORTD &= ~(1<< 5);}	
#define MIAC_LCD_RW_SET			{PORTD |= (1<< 5);}	
#define MIAC_LCD_RT_CLR			{PORTD &= ~(1<< 7);}	
#define MIAC_LCD_RT_SET			{PORTD |= (1<< 7);}	
#define MIAC_LCD_BL_CLR			{PORTG &= ~(1<< 2);}	
#define MIAC_LCD_BL_SET			{PORTG |= (1<< 2);}	

#define MIAC_LCD_EN_VAL(x)			{if (x) MIAC_LCD_EN_SET else MIAC_LCD_EN_CLR;}
#define MIAC_LCD_RS_VAL(x)			{if (x) MIAC_LCD_RS_SET else MIAC_LCD_RS_CLR;}
#define MIAC_LCD_RW_VAL(x)			{if (x) MIAC_LCD_RW_SET else MIAC_LCD_RW_CLR;}
#define MIAC_LCD_RT_VAL(x)			{if (x) MIAC_LCD_RT_SET else MIAC_LCD_RT_CLR;}
#define MIAC_LCD_BL_VAL(x)			{if (x) MIAC_LCD_BL_SET else MIAC_LCD_BL_CLR;}

#define MIAC_LCD_EN_VAL_DDR(x)		{DDRD |= (1<< 6); MIAC_LCD_EN_VAL(x);}	
#define MIAC_LCD_RS_VAL_DDR(x)		{DDRD |= (1<< 4); MIAC_LCD_RS_VAL(x);}	
#define MIAC_LCD_RW_VAL_DDR(x)		{DDRD |= (1<< 5); MIAC_LCD_RW_VAL(x);}	
#define MIAC_LCD_RT_VAL_DDR(x)		{DDRD |= (1<< 7); MIAC_LCD_RT_VAL(x);}	
#define MIAC_LCD_BL_VAL_DDR(x)		{DDRG |= (1<< 2); MIAC_LCD_BL_VAL(x);}	

#define MIAC_DB0_SET_DDR	{DDRC |= (1<<0); PORTC |= (1<<0);}
#define MIAC_DB1_SET_DDR	{DDRC |= (1<<1); PORTC |= (1<<1);}
#define MIAC_DB2_SET_DDR	{DDRC |= (1<<2); PORTC |= (1<<2);}
#define MIAC_DB3_SET_DDR	{DDRC |= (1<<3); PORTC |= (1<<3);}
#define MIAC_DB4_SET_DDR	{DDRC |= (1<<4); PORTC |= (1<<4);}
#define MIAC_DB5_SET_DDR	{DDRC |= (1<<5); PORTC |= (1<<5);}
#define MIAC_DB6_SET_DDR	{DDRC |= (1<<6); PORTC |= (1<<6);}
#define MIAC_DB7_SET_DDR	{DDRC |= (1<<7); PORTC |= (1<<7);}

#define MIAC_DB0_CLR_DDR	{DDRC |= (1<<0); PORTC &= ~(1<<0);}
#define MIAC_DB1_CLR_DDR	{DDRC |= (1<<1); PORTC &= ~(1<<1);}
#define MIAC_DB2_CLR_DDR	{DDRC |= (1<<2); PORTC &= ~(1<<2);}
#define MIAC_DB3_CLR_DDR	{DDRC |= (1<<3); PORTC &= ~(1<<3);}
#define MIAC_DB4_CLR_DDR	{DDRC |= (1<<4); PORTC &= ~(1<<4);}
#define MIAC_DB5_CLR_DDR	{DDRC |= (1<<5); PORTC &= ~(1<<5);}
#define MIAC_DB6_CLR_DDR	{DDRC |= (1<<6); PORTC &= ~(1<<6);}
#define MIAC_DB7_CLR_DDR	{DDRC |= (1<<7); PORTC &= ~(1<<7);}

#define MIAC_M1_SET_DDR	{DDRB |= (1<< 7); PORTB |= (1<< 7);}
#define MIAC_MA_SET_DDR	{DDRB |= (1<< 5); PORTB |= (1<< 5);}
#define MIAC_MB_SET_DDR	{DDRB |= (1<< 6); PORTB |= (1<< 6);}
#define MIAC_M2_SET_DDR	{DDRB |= (1<< 4); PORTB |= (1<< 4);}
#define MIAC_MC_SET_DDR	{DDRE |= (1<< 3); PORTE |= (1<< 3);}
#define MIAC_MD_SET_DDR	{DDRE |= (1<< 4); PORTE |= (1<< 4);}

#define MIAC_M1_CLR_DDR	{DDRB |= (1<< 7); PORTB &= ~(1<< 7);}
#define MIAC_MA_CLR_DDR	{DDRB |= (1<< 5); PORTB &= ~(1<< 5);}
#define MIAC_MB_CLR_DDR	{DDRB |= (1<< 6); PORTB &= ~(1<< 6);}
#define MIAC_M2_CLR_DDR	{DDRB |= (1<< 4); PORTB &= ~(1<< 4);}
#define MIAC_MC_CLR_DDR	{DDRE |= (1<< 3); PORTE &= ~(1<< 3);}
#define MIAC_MD_CLR_DDR	{DDRE |= (1<< 4); PORTE &= ~(1<< 4);}

#define MIAC_Q1_SET_DDR	{DDRA |= (1<<0); PORTA |= (1<<0);}
#define MIAC_Q2_SET_DDR	{DDRA |= (1<<1); PORTA |= (1<<1);}
#define MIAC_Q3_SET_DDR	{DDRA |= (1<<2); PORTA |= (1<<2);}
#define MIAC_Q4_SET_DDR	{DDRA |= (1<<3); PORTA |= (1<<3);}

#define MIAC_Q1_CLR_DDR	{DDRA |= (1<<0); PORTA &= ~(1<<0);}
#define MIAC_Q2_CLR_DDR	{DDRA |= (1<<1); PORTA &= ~(1<<1);}
#define MIAC_Q3_CLR_DDR	{DDRA |= (1<<2); PORTA &= ~(1<<2);}
#define MIAC_Q4_CLR_DDR	{DDRA |= (1<<3); PORTA &= ~(1<<3);}

#define MIAC_CS_CAN_CLR_DDR	{DDRE |= (1<< 6); PORTE &= ~(1<< 6);}
#define MIAC_CS_ADC_CLR_DDR	{DDRE |= (1<< 5); PORTE &= ~(1<< 5);}

#define MIAC_CS_CAN_SET_DDR	{DDRE |= (1<< 6); PORTE |= (1<< 6);}
#define MIAC_CS_ADC_SET_DDR	{DDRE |= (1<< 5); PORTE |= (1<< 5);}

#define WG_EDGE_DELAY	delayMicroseconds(10)
#define WG_EN_DELAY		delayMicroseconds(80)
#define WG_EN_FOLLOW	delayMicroseconds(10)
#define WG_BUSY_DELAY	delayMicroseconds(80)

miac::miac(void)
{
	// Setup LCD handshake outputs
	MIAC_LCD_EN_VAL_DDR(0);
	MIAC_LCD_RS_VAL_DDR(0);
	MIAC_LCD_RW_VAL_DDR(0);
	MIAC_LCD_RT_VAL_DDR(0);
	// Reset the gLCD
	delay_ms(250);
	MIAC_LCD_RT_VAL(1);
	delay_ms(100);
	// Initialise the gLCD
	WG_Command(0x30);		// 8 bit interface
	WG_Command(0x30);		// 8 bit interface
	WG_Command(0x01);		// clear RAM
	delay_ms(50);
	WG_Command(0xC0);		// cursor off, display on
	WG_Command(0xC0);		// cursor off, display on
	WG_Command(0x32);		// graphic mode on
	WG_Command(0x36);		// extended instruction set
	delay_ms(1);
	WG_Command(0x02);		// enable IRAM address
	WG_Command(0x40);		// IRAM 0
	WG_Command(0x80);		// GRAM 0
	WG_Command(0x80);		// GRAM 0
	delay_ms(100);
	clear();
	cursor(0,0);
	// Default gLCD backlight to On
	MIAC_LCD_BL_VAL_DDR(1);
	// Default text drawing mode to opaque (reset background pixels to off)
	WG_TextMode = 0;
}

void miac::clear(void)
{
	uint8_t row;
	uint8_t col;
	WG_WriteWordAt(0, 0, 0);
	for (row=0; row < 64; ++row)
	{
		for (col=0; col < 8; ++col)
			WG_WriteWordAt(0, col, row);
	}
	WG_CursorX = 0;
	WG_CursorY = 0;
}

void miac::cursor(uint8_t column, uint8_t row)
{
	WG_CursorX = column;
	WG_CursorY = row;
}

void miac::backlight(uint8_t state)
{
	MIAC_LCD_BL_VAL_DDR(state);
}

void miac::textMode(uint8_t mode)
{
	WG_TextMode = mode;
}

void miac::pixel(uint8_t px, uint8_t py, uint8_t state)
{
	// state = 0 turns off the pixel, else turns on (white)
	uint8_t BitRow = py + WG_ORIGIN_Y;
	uint8_t BitCol = px + WG_ORIGIN_X;
	uint8_t WordCol = BitCol / 16;
	uint8_t BitShift = BitCol % 16;
	uint16_t DisplayData = WG_ReadWordAt(WordCol, BitRow);

	if (state)
	{
		// draw the pixel
		DisplayData |= (0x8000 >> BitShift);
	}
	else
	{
		// delete the pixel
		DisplayData &= ~(0x8000 >> BitShift);
	}
	WG_WriteWordAt(DisplayData, WordCol, BitRow);	
}

void miac::line(uint8_t sx, uint8_t sy, uint8_t ex, uint8_t ey, uint8_t state)
{
	int16_t de, ox, oy, wx, wy;
	wx = ex - sx;
	wy = ey - sy;
	if ((wx>0?wx:-wx) > (wy>0?wy:-wy))
	{
		de = (sx > ex)?-1:1;
		for (ox=0; ox!=(wx+de); ox+=de)
		{
			oy = (ox * wy)/wx;
			pixel(sx + ox, sy + oy, state);
		}
	}
	else
	{
		de = (sy > ey)?-1:1;
		for (oy=0; oy!=(wy+de); oy+=de)
		{
			ox = (oy * wx)/wy;
			pixel(sx + ox, sy + oy, state);
		}
	}
}

void miac::rectangle(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t state)
{
	line(x1,y1,x2,y1, state);
	line(x2,y1,x2,y2, state);
	line(x2,y2,x1,y2, state);
	line(x1,y2,x1,y1, state);
}

void miac::glyph(uint8_t glyph, uint8_t px, uint8_t py, uint8_t mode)
{
	// The display row for the first row of the glyph ...
	uint8_t BitRowStart = py + WG_ORIGIN_Y;
	// The display column for the first col of the glyph
	uint8_t BitColStart = px + WG_ORIGIN_X;
	// The display word row start ...
	uint8_t WordCol = BitColStart / 16;
	// Bit shift for mask and glyph row ...
	uint8_t BitRemainder = BitColStart % 16;
	int8_t MaskShift = 16 - WG_CHAR_WIDTH - BitRemainder;
	
	uint8_t y;
	uint16_t GlyphRow;
	uint16_t DisplayData;
	
	for (y = 0; y < WG_CHAR_HEIGHT; ++y)
	{
		GlyphRow = pgm_read_byte(WG_CharMap + (glyph * WG_CHAR_HEIGHT) + y);
		
		if (MaskShift >= 0)
		{
			// glyph is fully contained within a single word
			DisplayData = WG_ReadWordAt(WordCol, BitRowStart + y);
			if (mode==0) DisplayData &= ~(WG_GLYPH_MASK << MaskShift);
			WG_WriteWordAt(DisplayData | (GlyphRow << MaskShift), WordCol, BitRowStart + y);
		}
		else
		{
			// glyph is not fully contained within a single word
			// do the first part
			DisplayData = WG_ReadWordAt(WordCol, BitRowStart + y);
			if (mode==0) DisplayData &= ~(WG_GLYPH_MASK >> (0-MaskShift));
			WG_WriteWordAt(DisplayData | (GlyphRow >> (0-MaskShift)), WordCol, BitRowStart + y);
			// now the second part in the next word
			DisplayData = WG_ReadWordAt(WordCol+1, BitRowStart + y);
			if (mode==0) DisplayData &= ~(WG_GLYPH_MASK << (16+MaskShift));
			WG_WriteWordAt(DisplayData | (GlyphRow << (16+MaskShift)), WordCol+1, BitRowStart + y);
		}
	}
}

void miac::outputHigh(uint8_t channel)
{
	switch (channel)
	{
		case 1: MIAC_M1_SET_DDR; MIAC_MA_SET_DDR; break;	// A
		case 2: MIAC_M1_SET_DDR; MIAC_MB_SET_DDR; break;	// B
		case 3: MIAC_M2_SET_DDR; MIAC_MC_SET_DDR; break;	// C
		case 4: MIAC_M2_SET_DDR; MIAC_MD_SET_DDR; break;	// D
	}
}

void miac::outputLow(uint8_t channel)
{
	switch (channel)
	{
		case 1: MIAC_M1_SET_DDR; MIAC_MA_CLR_DDR; break;	// A
		case 2: MIAC_M1_SET_DDR; MIAC_MB_CLR_DDR; break;	// B
		case 3: MIAC_M2_SET_DDR; MIAC_MC_CLR_DDR; break;	// C
		case 4: MIAC_M2_SET_DDR; MIAC_MD_CLR_DDR; break;	// D
	}
}

void miac::outputBridgePWM(uint8_t bridge, uint8_t duty, uint16_t prescaler)
{
	uint8_t clockSelect;
	// setup the prescaler
	if (prescaler < 4)
	{
		clockSelect = 1;	// /1
	}
	else if (prescaler < 32)
	{
		clockSelect = 2;	// /8
	}
	else if (prescaler < 128)
	{
		clockSelect = 3;	// /64
	}
	else if (prescaler < 512)
	{
		clockSelect = 4;	// /256
	}
	else
	{
		clockSelect = 5;	// /1024
	}
	
	switch (bridge)
	{
		case 1:		// A and B controlled by M1 on PB7
			TCCR1A &= ~0b00000100;
			TCCR1A |=  0b00001000;	// Clear OC1C on Compare Match, Mode 1 PWM
			TCCR1B = clockSelect;	// pre-scaler bits 1-5 = /1,/8,/64,/256,/1024
					// 
			break;
		case 2:		// C and D controlled by M2 on PB4
			TCCR2A &= ~0b01000000;
			TCCR2A |=  0b10000000;	// Clear OC2A on Compare Match, Mode 1 PWM
			TCCR2B = clockSelect;	// pre-scaler bits 1-5 = /1,/8,/64,/256,/1024
			break;
	}
	// and the duty
	outputBridgePWM(bridge, duty);
}

void miac::outputBridgePWM(uint8_t bridge, uint8_t duty)
{
	if ((duty == 0)||(duty == 255))
	{
		// disable PWM
		switch (bridge)
		{
			case 1:		// A and B controlled by M1 on PB7
				TCCR1A &= ~0b00001000;	
				break;
			case 2:		// C and D controlled by M2 on PB4
				TCCR2A &= ~0b10000000;
				break;
		}
	}
	else
	{
		// setup the duty
		switch (bridge)
		{
			case 1:		// A and B controlled by M1 on PB7
				OCR1C = duty;
				TCCR1A &= ~0b00000100;
				TCCR1A |=  0b00001000;	// OC1C on Compare Match, Mode 1 PWM			
				break;
			case 2:		// C and D controlled by M2 on PB4
				OCR2A = duty;
				TCCR2A &= ~0b01000000;
				TCCR2A |=  0b10000000;	// OC2A on Compare Match, Mode 1 PWM
				break;
		}
	}
}

void miac::relayOn(uint8_t relay)
{
	switch (relay)
	{
		case 1: MIAC_Q1_SET_DDR; break;	// Q1
		case 2: MIAC_Q2_SET_DDR; break;	// Q2
		case 3: MIAC_Q3_SET_DDR; break;	// Q3
		case 4: MIAC_Q4_SET_DDR; break;	// Q4
	}
}

void miac::relayOff(uint8_t relay)
{
	switch (relay)
	{
		case 1: MIAC_Q1_CLR_DDR; break;	// Q1
		case 2: MIAC_Q2_CLR_DDR; break;	// Q2
		case 3: MIAC_Q3_CLR_DDR; break;	// Q3
		case 4: MIAC_Q4_CLR_DDR; break;	// Q4
	}
}

uint8_t miac::digitalRead(uint8_t channel)
{
	switch(channel)
	{
		case 1:		MIAC_I1_IN_DDR; return (MIAC_I1_GET)?0:1;
		case 2:		MIAC_I2_IN_DDR; return (MIAC_I2_GET)?0:1;
		default:	return ((analogReadByte(channel) > 85)?1:0);
	}
}

uint8_t miac::analogReadByte(uint8_t channel)
{
	return analogReadWord(channel) >> 2;
}

uint16_t miac::analogReadWord(uint8_t channel)
{
	uint8_t buffer[3];
	uint8_t txn = 0;
	buffer[0] = 0x01;
	buffer[1] = 0x80 | ((channel-1) << 4);
	buffer[2] = 0;
											// set MOSI, SCK as output, others input
	DDRB &= ~(1<<3);						// MISO on B3 input
	DDRB |= (1<<2);							// MOSI on B2 output
	DDRB |= (1<<1);							// SCLK on B1 output
											// chip selects, all high (disabled)
	MIAC_CS_CAN_SET_DDR;					// CS_CAN on E6
	MIAC_CS_ADC_SET_DDR;					// CS_ADC on E5
											// Setup SPI Master mode, clock rate etc
	SPCR = (1 << MSTR);						// SPI Master mode (SPR1:SPR0=0 -> fosc/4)
#define MIAC_ADC_SPI_MODE	0
	if (MIAC_ADC_SPI_MODE & 0x01) SPCR |= (1 << CPHA);	// Clock Phase
	if (MIAC_ADC_SPI_MODE & 0x02) SPCR |= (1 << CPOL);	// Clock Polarity
	SPCR |= (1 << SPE);						// SPI Enable
	
	MIAC_CS_ADC_CLR_DDR;					// cs: 0 = CAN, 1 = ADC
	for (txn=0; txn<sizeof(buffer); ++txn)
	{
		SPDR = buffer[txn];					// send outgoing byte
		while (!(SPSR & (1<<SPIF)));		// wait until data has been sent
		buffer[txn] = SPDR;					// receive the byte
	}
	MIAC_CS_ADC_SET_DDR;
	return ((buffer[1] << 8) | buffer[2]) & 0x3ff;
}

uint8_t miac::keypadRead(void)
{
	uint8_t sample = 0;
	uint8_t strobe;

	MIAC_LCD_EN_CLR;
	MIAC_LCD_DB_IN_DDR;

	for (strobe=0; strobe<3; ++strobe)
	{
		switch(strobe)
		{
			case 0:
				// strobe 0
				MIAC_DB5_CLR_DDR;
				delayMicroseconds(10);
				sample = MIAC_LCD_DB_GET;
				MIAC_LCD_DB_IN_DDR;
				break;
			case 1:
				// strobe 1
				MIAC_DB4_CLR_DDR;
				delayMicroseconds(10);
				sample = MIAC_LCD_DB_GET;
				MIAC_LCD_DB_IN_DDR;
				break;
			case 2:
				// strobe 2
				MIAC_DB3_CLR_DDR;
				delayMicroseconds(10);
				sample = MIAC_LCD_DB_GET;
				MIAC_LCD_DB_IN_DDR;
				break;
		}
		if (!(sample & 0x01))
		{
			return (strobe * 3) + 0;
		}
		else if (!(sample & 0x02))
		{
			return (strobe * 3) + 1;
		}
		else if (!(sample & 0x04))
		{
			return (strobe * 3) + 2;
		}
	}
	return 255;
}

void miac::timeRead(char *time)
{
	uint8_t data;
	// Hours
	data = rtcDataRead(2);
	time[0] = ((data & 0x30) >> 4) + '0';
	time[1] = (data & 0x0f) + '0';
	// Minutes
	data = rtcDataRead(1);
	time[2] = ((data & 0x70) >> 4) + '0';
	time[3] = (data & 0x0f) + '0';
	// Seconds
	data = rtcDataRead(0);
	time[4] = ((data & 0x70) >> 4) + '0';
	time[5] = (data & 0x0f) + '0';
	time[6] = 0;
}

void miac::dateRead(char *date)
{
	uint8_t data;
	// Date
	data = rtcDataRead(4);
	date[0] = ((data & 0x30) >> 4) + '0';
	date[1] = (data & 0x0f) + '0';
	// Month
	data = rtcDataRead(5);
	date[2] = ((data & 0x10) >> 4) + '0';
	date[3] = (data & 0x0f) + '0';
	// Year
	data = rtcDataRead(6);
	date[4] = ((data & 0xf0) >> 4) + '0';
	date[5] = (data & 0x0f) + '0';
	date[6] = 0;
}

void miac::timeWrite(char *time)
{
	// Stop the clock
	rtcDataWrite(0, 0);
	// Hours
	rtcDataWrite(2, (((time[0] - '0') & 3) << 4) | ((time[1] - '0') & 0x0f));
	// Minutes
	rtcDataWrite(1, (((time[2] - '0') & 7) << 4) | ((time[3] - '0') & 0x0f));
	// Seconds and start the clock
	rtcDataWrite(0, (((time[4] - '0') & 7) << 4) | ((time[5] - '0') & 0x0f) | 0x80);
	// Enable battery backup
	rtcDataWrite(3, 8);
}

void miac::dateWrite(char *date)
{
	// Day
	rtcDataWrite(4, (((date[0] - '0') & 3) << 4) | ((date[1] - '0') & 0x0f));
	// Month
	rtcDataWrite(5, (((date[2] - '0') & 1) << 4) | ((date[3] - '0') & 0x0f));
	// Year
	rtcDataWrite(6, (((date[4] - '0') & 0x0f) << 4) | ((date[5] - '0') & 0x0f));
}

uint8_t miac::rtcDataRead(uint8_t address)
{
	uint8_t data;
	i2cStart();
	i2cTxByte(0xDE);
	i2cTxByte(address);
	i2cRestart();
	i2cTxByte(0xDF);
	data = i2cRxByte(1);
	i2cStop();
	return (data);
}

void miac::rtcDataWrite(uint8_t address, uint8_t data)
{
	i2cStart();
	i2cTxByte(0xDE);
	i2cTxByte(address);
	i2cTxByte(data);
	i2cStop();
}

uint8_t miac::eepromRead(uint16_t address)
{
	uint8_t data;
	i2cStart();
	i2cTxByte(0xA0);
	i2cTxByte(address>>8);
	i2cTxByte(address);
	i2cRestart();
	i2cTxByte(0xA1);
	data = i2cRxByte(1);
	i2cStop();
	return (data);
}

void miac::eepromWrite(uint16_t address, uint8_t data)
{
	i2cStart();
	i2cTxByte(0xA0);
	i2cTxByte(address>>8);
	i2cTxByte(address);
	i2cTxByte(data);
	i2cStop();
}

void miac::i2cStart(void)
{
	#define MX_CLOCK_SPEED 8000000
	#define MX_RTC_BAUD 100000
	TWBR = (((MX_CLOCK_SPEED / MX_RTC_BAUD)-16)/2);
	TWSR = 0;
	TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
	while((TWCR & (1 << TWINT)) == 0);
}

uint8_t miac::i2cTxByte(uint8_t data)
{
	TWDR = data;
	delayMicroseconds(10);
	TWCR = (1 << TWINT) | (1 << TWEN);
	while((TWCR & (1 << TWINT)) == 0);
	if(((TWSR & 0xf8) == 0x18) || ((TWSR & 0xf8) == 0x28))
		return(0);
	else
		return(1);
}

uint8_t miac::i2cRxByte(uint8_t last)
{
	if (last)
		TWCR = (1 << TWINT) | (1 << TWEN);
	else
		TWCR = (1 << TWINT) | (1 << TWEA) | (1 << TWEN);
	while((TWCR & (1 << TWINT)) == 0);
	return(TWDR);
}

void miac::i2cRestart(void)
{
	TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
	while((TWCR & (1 << TWINT)) == 0);
}

void miac::i2cStop(void)
{
	TWCR = (1 << TWINT) | (1 << TWSTO) | (1 << TWEN);
	delay_ms(10);
}

size_t miac::write(uint8_t c)
{
	switch(c)
	{
	case 0:
		return 0;
		
	case '\n':	// 0x0a newline
		WG_CursorX = 0;
		if (++WG_CursorY >= WG_CHAR_ROWS)
		{
			WG_CursorY = 0;
		}
		break;
		
	case '\r':	// 0x0d carriage return
		WG_CursorX = 0;
		break;

	case '\f':	// 0x0c Form feed, new page
		clear();
		break;

	default:
		if (c >= WG_CHAR_COUNT)
		{
			switch (c)
		{
				case 0xB0:
				case 0xDF:
				c = 0x1F;	// re-map the Deg symbol
					break;
				case 0xC2:		// ignore esc
				default:		// or anything else
					return 1;
			}
		}
			// cursor boundary checks ...
			if (WG_CursorY >= WG_CHAR_ROWS) WG_CursorY = 0;
			if (WG_CursorX >= WG_CHAR_COLS) WG_CursorX = 0;

			glyph(c, (WG_CursorX * WG_CHAR_WIDTH), (WG_CursorY * WG_CHAR_HEIGHT), WG_TextMode);
			// move the cursor ...
			if (++WG_CursorX >= WG_CHAR_COLS)
			{
				WG_CursorX = 0;
				if (++WG_CursorY >= WG_CHAR_ROWS)
				{
					WG_CursorY = 0;
			}
		}
		break;
	}
	return 1;
}

// write a data byte to DB0-7
void miac::WG_Write(uint8_t data)
{
	MIAC_LCD_RW_VAL(0);				// RW low
	MIAC_LCD_RS_VAL(1);				// RS high
	// setup the DB0-7 data bits
	// set DB0-7 to output
	MIAC_LCD_DB_VAL_DDR(data);
	// toggle EN
	MIAC_LCD_EN_VAL(1);
	WG_EN_DELAY;
	MIAC_LCD_EN_VAL(0);
	WG_EN_FOLLOW;
}

// read a status byte from DB0-7
uint8_t miac::WG_Status(void)
{
	uint8_t status;
	// set DB0-7 to input
	MIAC_LCD_DB_IN_DDR;
	MIAC_LCD_RS_VAL(0);
	MIAC_LCD_RW_VAL(1);
	WG_EDGE_DELAY;
	MIAC_LCD_EN_VAL(1);
	WG_EN_DELAY;
	// read the DB0-7 data bits
	status = MIAC_LCD_DB_GET;
	MIAC_LCD_EN_VAL(0);
	WG_EN_FOLLOW;
	return status;
}

// read a data byte from DB0-7
uint8_t miac::WG_Read(void)
{
	uint8_t data;
	// set DB0-7 to input
	MIAC_LCD_DB_IN_DDR;
	MIAC_LCD_RS_VAL(1);
	MIAC_LCD_RW_VAL(1);
	WG_EDGE_DELAY;
	MIAC_LCD_EN_VAL(1);
	WG_EN_DELAY;
	// read the DB0-7 data bits
	data = MIAC_LCD_DB_GET;
	MIAC_LCD_EN_VAL(0);
	WG_EN_FOLLOW;
	return data;
}

// write a command byte to DB0-7 and do the handshake
void miac::WG_Command(uint8_t cmd)
{
	MIAC_LCD_RW_VAL(0);				// RW low
	MIAC_LCD_RS_VAL(0);				// RS low
	// setup the DB0-7 data bits
	// set DB0-7 to output
	MIAC_LCD_DB_VAL_DDR(cmd);				
	MIAC_LCD_EN_VAL(1);
	WG_EN_DELAY;
	MIAC_LCD_EN_VAL(0);
	WG_EN_FOLLOW;
}

void miac::WG_WriteWordAt(uint16_t w, uint8_t x, uint8_t y)
{
	uint16_t temp = w;
	uint8_t i = 16;
	
	// sanity check on parameters
	if ((x > 7)||(y > 63)) return;

	// rotate the screen data
	// re-config the memory map
	if (y > 31)
	{
		y = 31 - y - 32;
		x = 7 - x;
	}
	else
	{
		y = 31 - y;
		x = 7 - x + 8;
	}
	w = 0;
	while(i--)
	{
		w <<= 1;
		if (temp & 1) w |= 1;
		temp >>= 1;
	}
	// wait for Busy flag inactive
	while (WG_Status() & 0x80);
	// 1.   Set vertical address ( Y ) for GDRAM 
	WG_Command(0x80 | (y & 0x3f));
	WG_EDGE_DELAY;
	// 2.   Set horizontal address ( X ) for GDRAM 
	WG_Command(0x80 | (x & 0x0f));
	WG_EDGE_DELAY;
	// 3.   Write D15 - D8 to GDRAM (first byte) 
	WG_Write(w >> 8);
	WG_EDGE_DELAY;
	// 4.   Write D7 - D0 to GDRAM (second byte)
	WG_Write(w);
	WG_EDGE_DELAY;
}

uint16_t miac::WG_ReadWordAt(uint8_t x, uint8_t y)
{
	uint16_t temp;
	uint16_t w;
	uint8_t i = 16;
	
	// sanity check on parameters
	if ((x > 7)||(y > 63)) return 0;
	
	// re-config the memory map
	if (y > 31)
	{
		y = 31 - y - 32;
		x = 7 - x;
	}
	else
	{
		y = 31 - y;
		x = 7 - x + 8;
	}
	// wait for Busy flag inactive
	while (WG_Status() & 0x80);
	// 1.   Set vertical address ( Y ) for GDRAM 
	WG_Command(0x80 | (y & 0x3f));
	WG_EDGE_DELAY;
	// 2.   Set horizontal address ( X ) for GDRAM 
	WG_Command(0x80 | (x & 0x0f));
	// dummy read
	WG_BUSY_DELAY;
	WG_Read();
	WG_BUSY_DELAY;
	temp = WG_Read() << 8;
	WG_BUSY_DELAY;
	temp |= WG_Read();
	WG_EDGE_DELAY;
	// mirror the screen data
	w = 0;
	while(i--)
	{
		w <<= 1;
		if (temp & 1) w |= 1;
		temp >>= 1;
	}
	return w;
}

void miac::delay_ms(uint16_t ms)
{
	while (ms--) delayMicroseconds(1000);
}

// CAN ...

#define MIAC_CS_CAN_CLR_DDR	{DDRE |= (1<< 6); PORTE &= ~(1<< 6);}
#define MIAC_CS_ADC_CLR_DDR	{DDRE |= (1<< 5); PORTE &= ~(1<< 5);}

#define MIAC_CS_CAN_SET_DDR	{DDRE |= (1<< 6); PORTE |= (1<< 6);}
#define MIAC_CS_ADC_SET_DDR	{DDRE |= (1<< 5); PORTE |= (1<< 5);}

#define CAN_REQOP_CONFIG	(0x80)
#define CAN_REQOP_LISTEN	(0x60)
#define CAN_REQOP_LOOPBACK	(0x40)
#define CAN_REQOP_NORMAL	(0x00)
#define CAN_REQOP_SLEEP		(0x20)
#define CAN_CANCTRL			(0x0f)
#define CAN_CANSTAT			(0x0e)
#define CAN_WRITE			(0x02)
#define CAN_READ			(0x03)
#define CAN_CNF1			(0x2a)
#define CAN_CNF2			(0x29)
#define CAN_CNF3			(0x28)
#define CAN_TXRTSCTRL		(0x0d)
#define CAN_EFLG			(0x2d)
#define CAN_TXB0CTRL		(0x30)
#define CAN_TXB1CTRL		(0x40)
#define CAN_TXB2CTRL		(0x50)
#define CAN_RX_STATUS		(0xb0)
#define CAN_RXB0CTRL		(0x60)
#define CAN_RXB1CTRL		(0x70)


void miac::transaction(uint8_t* buffer, uint8_t count)
{
	uint8_t txn = 0;
											// set MOSI, SCK as output, others input
	DDRB &= ~(1<<3);						// MISO on B3 input
	DDRB |= (1<<2);							// MOSI on B2 output
	DDRB |= (1<<1);							// SCLK on B1 output
											// chip selects, all high (disabled)
	MIAC_CS_CAN_SET_DDR;					// CS_CAN on E6
	MIAC_CS_ADC_SET_DDR;					// CS_ADC on E5
											// Setup SPI Master mode, clock rate etc
	SPCR = (1 << MSTR);						// SPI Master mode (SPR1:SPR0=0 -> fosc/4)
#define MIAC_CAN_SPI_MODE	0
	if (MIAC_CAN_SPI_MODE & 0x01) SPCR |= (1 << CPHA);	// Clock Phase
	if (MIAC_CAN_SPI_MODE & 0x02) SPCR |= (1 << CPOL);	// Clock Polarity
	SPCR |= (1 << SPE);						// SPI Enable
	
	MIAC_CS_CAN_CLR_DDR;
	while (count--)
	{
		SPDR = buffer[txn];					// send outgoing byte
		while (!(SPSR & (1<<SPIF)));		// wait until data has been sent
		buffer[txn++] = SPDR;				// receive the byte
	}
	MIAC_CS_CAN_SET_DDR;
}

void miac::command(uint8_t newCommand)
{
	uint8_t buffer[1];
	buffer[0] = newCommand;
	transaction(buffer, sizeof(buffer));
}

void miac::reset(void)
{
	command(0xc0);
}

void miac::registerWrite(uint8_t address, uint8_t data)
{
	uint8_t buffer[3];
	buffer[0] = CAN_WRITE;
	buffer[1] = address;
	buffer[2] = data;
	transaction(buffer, sizeof(buffer));
}

uint8_t miac::registerRead(uint8_t address)
{
	uint8_t buffer[3];
	buffer[0] = CAN_READ;
	buffer[1] = address;
	transaction(buffer, sizeof(buffer));
	return buffer[2];
}

void miac::config(uint8_t newConfig)
{
	uint8_t mode;
	uint8_t timeout = 0;
	do
	{
		registerWrite(CAN_CANCTRL, newConfig);
		delayMicroseconds(1000);
		mode = registerRead(CAN_CANSTAT);
		delayMicroseconds(1000);
	} while (((mode & 0xe0)!= newConfig) && (++timeout < 128));
}

uint32_t miac::unpackID(uint8_t exid, uint8_t *buffer)
{
	uint32_t ident;
	if (exid)
	{
		// Extended ID
		ident = ((uint32_t)((buffer[0])<<3 | (buffer[1]>>5))<<18) | ((uint32_t)(buffer[1] & 3)<<16) | (buffer[2]<<8) | (buffer[3]);
	}
	else
	{
		// Standard ID
		ident = (buffer[0])<<3 | (buffer[1]>>5);
	}
	return ident;
}

void miac::packID(uint8_t exid, uint32_t ident, uint8_t *buffer)
{
	if (exid)
	{
		// Extended ID
		buffer[0] = ident >> 21;
		buffer[1] = ((ident >> 16)&3) | ((ident >> 13)&0xe0) | 8;
		buffer[2] = ident >> 8;
		buffer[3] = ident;
	}
	else
	{
		// Standard ID
		buffer[0] = ident >> 3;
		buffer[1] = ident << 5;
	}
}


/**********************************************
 CAN public functions
**********************************************/
uint8_t miac::rxStatusRead(void)
{
	uint8_t buffer[3];
	buffer[0] = CAN_RX_STATUS;
	transaction(buffer, sizeof(buffer));
	return buffer[1];
}

uint32_t miac::rxIdentRead(uint8_t rxBuffer)
{
	return msgID[rxBuffer & 1];
}

uint8_t miac::rxDataCountRead(uint8_t rxBuffer)
{
	return msgCount[rxBuffer & 1];
}

uint8_t miac::rxDataRead(uint8_t rxBuffer, uint8_t index)
{
	return msgData[rxBuffer & 1][index & 7];
}

void miac::txSend(uint8_t txBuffer)
{
	switch(txBuffer)
	{
		case 0:	command(0x81); break;
		case 1:	command(0x82); break;
		case 2:	command(0x84); break;
	}
}

miac::miac(uint32_t busRate)
{
	miac();
	reset();
	config(CAN_REQOP_CONFIG);
	switch(busRate)
	{
		case 125:
		case 125000:
			registerWrite(CAN_CNF1, 7);
			break;
		case 250:
		case 250000:
			registerWrite(CAN_CNF1, 3);
			break;
		case 500:
		case 500000:
			registerWrite(CAN_CNF1, 1);
			break;
		default:
			registerWrite(CAN_CNF1, 0);
			break;
	}
	registerWrite(CAN_CNF2, 0xc9);		// sample at 50%
	registerWrite(CAN_CNF3, 4);
	registerWrite(CAN_TXRTSCTRL, 0);
	registerWrite(CAN_EFLG, 0);
	registerWrite(CAN_CANCTRL, 0x80);
	config(CAN_REQOP_NORMAL);
}

uint8_t miac::rxCheck(uint8_t rxBuffer)
{
	uint8_t buffer[14];
	uint8_t x;
	uint8_t status = rxStatusRead();
	if (((rxBuffer == 0)&&(status & 0x40))||((rxBuffer == 1)&&(status & 0x80)))
	{
		buffer[0] = rxBuffer?0x94:0x90;
		transaction(buffer, sizeof(buffer));
		
		// message ID (depends on EXID flag)
		if (buffer[2] & 0x08)
		{
			// Extended ID
			msgEXID[rxBuffer] = 1;
			msgID[rxBuffer] = unpackID(1, &buffer[1]);
		}
		else
		{
			// Standard ID
			msgEXID[rxBuffer] = 0;
			msgID[rxBuffer] = unpackID(0, &buffer[1]);
		}
		// message DLC
		msgCount[rxBuffer] = buffer[5] & 0x0f;
		// message data
		for (x=0; x<8; ++x)
		{
			msgData[rxBuffer][x] = buffer[x+6];
		}
		return 1;
	}
	else
	{
		return 0;
	}
}

uint8_t miac::txDataWrite(uint8_t txBuffer, uint8_t count, uint8_t *data)
{
	uint8_t buffer[10];
	if ((txBuffer<3)&&(count<9))
	{
		buffer[0] = CAN_WRITE;
		buffer[1] = 0x35 + (txBuffer * 0x10);
		buffer[2] = count;
		while (count--)
			buffer[3+count] = data[count];
		transaction(buffer, sizeof(buffer));
		return 1;
	}
	return 0;
}

uint8_t miac::txDataWrite(uint8_t txBuffer, uint8_t count, char *data)
{
	return txDataWrite(txBuffer, count, (uint8_t *)data);
}

uint8_t miac::txDataWrite(uint8_t txBuffer, uint8_t count, const char *data)
{
	return txDataWrite(txBuffer, count, (uint8_t *)data);
}

uint8_t miac::txIdentWrite(uint8_t txBuffer, uint8_t exid, uint32_t ident)
{
	uint8_t buffer[5];
	if (txBuffer > 2)
		return 0;
	buffer[0] = 0x40 + (txBuffer * 2);
	packID(exid, ident, &buffer[1]);
	transaction(buffer, sizeof(buffer));
	return 1;
}

// mask and filter functions
uint8_t miac::rxMaskWrite(uint8_t rxMaskNo, uint8_t exid, uint32_t mask)
{
	uint8_t buffer[6];
	if (rxMaskNo > 1)
		return 0;
	buffer[0] = CAN_WRITE;
	buffer[1] = 0x20 + (rxMaskNo * 4);
	packID(exid, mask, &buffer[2]);
	transaction(buffer, sizeof(buffer));
	return 1;
}

uint8_t miac::rxFilterWrite(uint8_t rxFilterNo, uint8_t exid, uint32_t filter)
{
	uint8_t buffer[6];
	if (rxFilterNo > 5)
		return 0;
	buffer[0] = CAN_WRITE;
	buffer[1] = (rxFilterNo<3)?(rxFilterNo*4):((rxFilterNo*4)+4);
	packID(exid, filter, &buffer[2]);
	transaction(buffer, sizeof(buffer));
	return 1;
}





