
/*
  miac.h - Library driver for the Matrix MIAC Industrial Controller
  Created by Leigh Morris (Matrix TSL) March 2016

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
*/
#ifndef miac_h
#define miac_h

#include "Arduino.h"
	
class miac : public Print
{
	public:
		miac(void);
		/* MIAC intialiser without CAN support
		 */
		void backlight(uint8_t state);
		/* Controls the gLCD backlight
		 * state		1=on, 0=off
		 */
		void clear(void);
		/* Clears the gLCD screen
		 */
		void cursor(uint8_t column, uint8_t row);
		/* Places the gLCD text cursor at column and row
		 * column		0 to 19 (when using default font)
		 * row			0 to 4 (when using default font)
		 */
		void pixel(uint8_t px, uint8_t py, uint8_t state);
		/* Draws or clears a pixel at px:py
		 * px			pixel column relative to the drawing origin (top left)
		 * py			pixel row relative to the drawing origin (top left)
		 * state		0=clear pixel, 1=active pixel
		 */
		void line(uint8_t sx, uint8_t sy, uint8_t ex, uint8_t ey, uint8_t state);
		/* Draws or clears a line from sx:sy to ex:ey
		 * sx			start column position relative to the drawing origin (top left)
		 * sy			start row position relative to the drawing origin (top left)
		 * ex			end column position relative to the drawing origin (top left)
		 * ey			end row position relative to the drawing origin (top left)
		 * state		0=clear pixel, 1=active pixel
		 */
		 void rectangle(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t state);
		/* Draws or clears a rectangle from x1:y1 to x2:y2
		 * x1			first corner column position relative to the drawing origin (top left)
		 * y1			first corner row position relative to the drawing origin (top left)
		 * x2			second corner column position relative to the drawing origin (top left)
		 * y2			second corner row position relative to the drawing origin (top left)
		 * state		0=clear pixel, 1=active pixel
		 */
		void glyph(uint8_t glyph, uint8_t px, uint8_t py, uint8_t mode);
		/* Draws a character glyph at point px:py
		 * glyph		character glyph to draw from font table
		 * px			pixel column relative to the drawing origin (top left)
		 * py			pixel row relative to the drawing origin (top left)
		 * mode			0=opaque/cleared background, 1=transparent background
		 */
		void textMode(uint8_t mode);
		/* Sets the text drawing mode for future text operations
		 * mode			0=opaque/cleared background, 1=transparent background
		 */
		void outputHigh(uint8_t channel);
		/* Sets a solid state output channel to high (current source)
		 * channel		1=A, 2=B, 3=C, 4=D
		 */
		void outputLow(uint8_t channel);
		/* Sets a solid state output channel to low (current sink)
		 * channel		1=A, 2=B, 3=C, 4=D
		 */
		void outputBridgePWM(uint8_t bridge, uint8_t duty, uint16_t prescaler);
		/* Sets up and enables PWM output on one of the two full bridges
		 * bridge		1=Bridge on A:B, 2=Bridge on C:D
		 * duty			duty cycle value 1 to 254 (note that 0 and 255 turn off PWM)
		 * prescaler	clock prescaler divider value (1, 8, 64, 256, 1024)
		 */
		void outputBridgePWM(uint8_t bridge, uint8_t duty);
		/* Sets or changes the PWM duty cycle on one of the two full bridges
		 * bridge		1=Bridge on A:B, 2=Bridge on C:D
		 * duty			duty cycle value 1 to 254 (note that 0 and 255 turn off PWM)
		 */
		void relayOn(uint8_t relay);
		/* Switches on one of the four relay outputs
		 * relay		1=Q1, 2=Q2, 3=Q3, 4=Q4
		 */
		void relayOff(uint8_t relay);
		/* Switches off one of the four relay outputs
		 * relay		1=Q1, 2=Q2, 3=Q3, 4=Q4
		 */
		uint8_t digitalRead(uint8_t channel);
		/* Returns the digital state of an input channel
		 * channel		1 to 8 for inputs I1 to I8
		 * returns		0=input low, 1=input high
		 */
		uint8_t analogReadByte(uint8_t channel);
		/* Returns the analogue value (8 bit) of an input channel
		 * channel		1 to 8 for inputs I1 to I8
		 * returns		8 bit analogue value
		 */
		uint16_t analogReadWord(uint8_t channel);
		/* Returns the analogue value (10 bit) of an input channel
		 * channel		1 to 8 for inputs I1 to I8
		 * returns		10 bit analogue value
		 */
		uint8_t keypadRead(void);
		/* Returns the Keypad state
		 * returns		0 to 8 key value or 255 if no key pressed
		 */

		void timeRead(char *time);
		/* Reads the time as a string HHMMSS from the RTC
		 * time			pointer to buffer to receive the string, must be at least 7 bytes
		 */

		void dateRead(char *date);
		/* Reads the date as a string DDMMYY from the RTC
		 * date			pointer to buffer to receive the string, must be at least 7 bytes
		 */

		void timeWrite(char *time);
		/* Writes the time from a string HHMMSS to the RTC
		 * time			pointer to buffer containing the string
		 */

		void dateWrite(char *date);
		/* Writes the date from a string DDMMYY to the RTC
		 * date			pointer to buffer containing the string
		 */

		uint8_t rtcDataRead(uint8_t address);
		/* Reads a data byte from the RTC device
		 * address		Address of the location to be read
		 * returns		The data from the requested address location
		 */

		void rtcDataWrite(uint8_t address, uint8_t data);
		/* Writes a data byte to the RTC device
		 * address		Address of the location to be written
		 * data			The data to be written to the device location
		 */

		uint8_t eepromRead(uint16_t address);
		/* Reads a data byte from the EEPROM device
		 * address		Address of the location to be read
		 * returns		The data from the requested address location
		 */

		void eepromWrite(uint16_t address, uint8_t data);
		/* Writes a data byte to the EEPROM device
		 * address		Address of the location to be written
		 * data			The data to be written to the device location
		 */
		 
		/************************************************************
		 * CAN bus support
		 ***********************************************************/
		
		miac(uint32_t busRate);
		/* MIAC intialiser and CAN setup
		 * busRate		125, 250, 500, 1000
		 */
		uint8_t txIdentWrite(uint8_t txBuffer, uint8_t exid, uint32_t ident);
		/* Writes the CAN identifier to a transmit buffer
		 * txBuffer		buffer to set (0, 1, 2)
		 * exid			0=standard identifier mode, 1=extended identifier mode
		 * ident		CAN Identifier code to set
		 * returns		1=successful, 0=error (invalid parameters)
		 */
		uint8_t txDataWrite(uint8_t txBuffer, uint8_t count, uint8_t *data);
		uint8_t txDataWrite(uint8_t txBuffer, uint8_t count, char *data);
		uint8_t txDataWrite(uint8_t txBuffer, uint8_t count, const char *data);
		/* Writes the packet data for a transmit buffer
		 * txBuffer		buffer to set (0, 1, 2)
		 * count		number of data bytes to write (1 to 8)
		 * data			data array
		 * returns		1=successful, 0=error (invalid parameters)
		 */
		void txSend(uint8_t txBuffer);		
		/* Initiates sending of a transmit buffer
		 * txBuffer		buffer to send (0, 1, 2)
		 */
		uint8_t rxMaskWrite(uint8_t rxMaskNo, uint8_t exid, uint32_t mask);
		/* Writes an identifier mask
		 * rxMaskNo		mask to set (0 or 1)
		 * exid			0=standard identifier mode, 1=extended identifier mode
		 * mask			mask value to set
		 * returns		1=successful, 0=error (invalid parameters)
		 */
		uint8_t rxFilterWrite(uint8_t rxFilterNo, uint8_t exid, uint32_t filter);
		/* Writes an identifier filter
		 * rxFilterNo	filter to set (0 to 5)
		 * exid			0=standard identifier mode, 1=extended identifier mode
		 * filter		filter value to set
		 * returns		1=successful, 0=error (invalid parameters)
		 */
		uint8_t rxStatusRead(void);
		/* Returns the status register of the controller
		 */
		uint8_t rxCheck(uint8_t rxBuffer);
		/* Checks for received message for the given buffer and retrieves the data
		 * txBuffer		buffer to check (0 or 1)
		 * returns		0=no message, 1=message received
		 */
		uint32_t rxIdentRead(uint8_t rxBuffer);
		/* Returns the CAN message identifier for the last received message for the given buffer
		 * txBuffer		buffer to read (0 or 1)
		 * returns		CAN message identifier
		 */
		uint8_t rxDataCountRead(uint8_t rxBuffer);
		/* Returns the CAN message data count for the last received message for the given buffer
		 * txBuffer		buffer to read (0 or 1)
		 * returns		message data byte count (size)
		 */
		uint8_t rxDataRead(uint8_t rxBuffer, uint8_t index);
		/* Returns a CAN message data byte for the last received message for the given buffer
		 * txBuffer		buffer to read (0 or 1)
		 * index		index of the message byte to read (index 0 to 7)
		 * returns		message data byte
		 */

	private:
		uint8_t WG_CursorX;
		uint8_t WG_CursorY;
		uint8_t WG_TextMode;
		
		void WG_Write(uint8_t data);
		uint8_t WG_Read(void);
		uint8_t WG_Status(void);
		void WG_Command(uint8_t cmd);
		void WG_WriteWordAt(uint16_t w, uint8_t x, uint8_t y);
		uint16_t WG_ReadWordAt(uint8_t x, uint8_t y);
		size_t write(uint8_t c);
		void delay_ms(uint16_t ms);
		void i2cStart(void);
		uint8_t i2cTxByte(uint8_t data);
		uint8_t i2cRxByte(uint8_t last);
		void i2cRestart(void);
		void i2cStop(void);
		// CAN
		uint32_t msgID[2];
		uint8_t msgCount[2];
		uint8_t msgData[2][8];
		uint8_t msgEXID[2];
		void transaction(uint8_t* buffer, uint8_t count);
		void command(uint8_t newCommand);
		void reset(void);
		void registerWrite(uint8_t address, uint8_t data);
		uint8_t registerRead(uint8_t address);
		void config(uint8_t newConfig);
		uint32_t unpackID(uint8_t exid, uint8_t *buffer);
		void packID(uint8_t exid, uint32_t ident, uint8_t *buffer);

};

#endif

